# © Copyright Databand.ai, an IBM Company 2022

import logging

from py4j.protocol import Py4JError, Py4JJavaError
from pydeequ.repository import MetricsRepository


logger = logging.getLogger(__name__)


class DbndMetricsRepository(MetricsRepository):
    """
    Deequ metrics repository implementation which submits all metrics to the Databand.
    It will proxy all calls to the Java implementation.
    This repository will work only if dbnd-api-deequ jar was added to the classpath.
    Otherwise, Deequ metrics won't be reported.

    Former implementation was used Java converter to convert Deequ metrics to dict and submit it
    via python log_metric API.
    However, that implementation doesn't work on Dataproc (it requires back channel from java to python),
    so we've switched to direct reporting from Java code.
    Task context are managed via set_external_context method which instructs JVM where to send metrics.
    """

    def __init__(self, spark_session):
        self._spark_session = spark_session
        self._jvm = spark_session._jvm
        self._jspark_session = spark_session._jsparkSession
        try:
            self.deequDBNDmetRep = (
                spark_session._jvm.ai.databand.deequ.DbndMetricsRepository
            )
            self.repository = self.deequDBNDmetRep()
        except (Py4JJavaError, Py4JError, TypeError):
            logger.error(
                "Unable to load Databand Java SDK. Deequ metrics won't be reported to Databand. "
                "Perhaps you forgot to add dbnd-api-deequ to spark classpath? "
                "Built-in Deequ in-memory metrics repository will be used instead."
            )
            self.deequIMmetRep = (
                spark_session._jvm.com.amazon.deequ.repository.memory.InMemoryMetricsRepository
            )
            self.repository = self.deequIMmetRep()



# © Copyright Databand.ai, an IBM Company 2022

import pyspark

from dbnd_spark.spark_targets.spark_marshalling import (
    SparkDataFrameToCsv,
    SparkMarshaller,
)
from dbnd_spark.spark_targets.spark_values import SparkDataFrameValueType
from targets.marshalling import register_marshaller
from targets.target_config import FileFormat
from targets.values import register_value_type


def dbnd_register_spark_types():
    register_value_type(SparkDataFrameValueType())

    for file_format, marshaller in (
        (FileFormat.txt, SparkMarshaller(fmt=FileFormat.txt)),
        (FileFormat.csv, SparkDataFrameToCsv()),
        (FileFormat.json, SparkMarshaller(fmt=FileFormat.json)),
        (FileFormat.parquet, SparkMarshaller(fmt=FileFormat.parquet)),
    ):
        register_marshaller(pyspark.sql.DataFrame, file_format, marshaller)



# © Copyright Databand.ai, an IBM Company 2022

import os
import re

from io import StringIO
from logging import StreamHandler

from dbnd._core.errors.friendly_error.task_execution import (
    failed_spark_status,
    failed_to_run_spark_script,
)
from dbnd._core.plugin.dbnd_plugins import is_airflow_enabled
from dbnd._core.utils.structures import list_of_strings
from dbnd_spark._core.spark_error_parser import parse_spark_log_safe
from dbnd_spark.local.local_spark_config import SparkLocalEngineConfig
from dbnd_spark.spark_ctrl import SparkCtrl


class LocalSparkExecutionCtrl(SparkCtrl):
    def run_pyspark(self, pyspark_script):
        jars = list(self.config.jars)
        application = pyspark_script
        if self.config.main_jar:
            jars += [self.config.main_jar]

        return self._run_spark_submit(application=application, jars=jars)

    def run_spark(self, main_class):
        application = self.config.main_jar
        return self._run_spark_submit(application=application, jars=self.config.jars)

    def _run_spark_submit(self, application, jars):
        # task_env = get_cloud_config(Clouds.local)
        spark_local_config = SparkLocalEngineConfig()
        _config = self.config
        deploy = self.deploy

        AIRFLOW_ON = is_airflow_enabled()

        if AIRFLOW_ON:
            from airflow.contrib.hooks.spark_submit_hook import SparkSubmitHook
            from airflow.exceptions import AirflowException as SparkException
        else:
            from dbnd_spark._vendor.airflow.spark_hook import (
                SparkException,
                SparkSubmitHook,
            )

        spark = SparkSubmitHook(
            conf=_config.conf,
            conn_id=spark_local_config.conn_id,
            name=self.job.job_id,
            application_args=list_of_strings(self.task.application_args()),
            java_class=self.task.main_class,
            files=deploy.arg_files(_config.files),
            py_files=deploy.arg_files(self.task.get_py_files()),
            driver_class_path=_config.driver_class_path,
            jars=deploy.arg_files(jars),
            packages=_config.packages,
            exclude_packages=_config.exclude_packages,
            repositories=_config.repositories,
            total_executor_cores=_config.total_executor_cores,
            executor_cores=_config.executor_cores,
            executor_memory=_config.executor_memory,
            driver_memory=_config.driver_memory,
            keytab=_config.keytab,
            principal=_config.principal,
            num_executors=_config.num_executors,
            env_vars=self._get_env_vars(),
            verbose=_config.verbose,
            status_poll_interval=_config.status_poll_interval,
        )
        if not AIRFLOW_ON:
            # If there's no Airflow then there's no Connection so we
            # take conn information from spark config
            spark.set_connection(spark_local_config.conn_uri)

        log_buffer = StringIO()
        with log_buffer as lb:
            dbnd_log_handler = self._capture_submit_log(spark, lb)
            try:
                # sync the application file to remote if needed
                spark.submit(application=deploy.sync(application))
            except SparkException as ex:
                return_code = self._get_spark_return_code_from_exception(ex)
                if return_code != "0":
                    error_snippets = parse_spark_log_safe(
                        log_buffer.getvalue().split(os.linesep)
                    )
                    raise failed_to_run_spark_script(
                        self,
                        spark._build_spark_submit_command(application=application),
                        application,
                        return_code,
                        error_snippets,
                    )
                else:
                    raise failed_spark_status(ex)
            finally:
                spark.log.handlers = [
                    h for h in spark.log.handlers if not dbnd_log_handler
                ]

    def _capture_submit_log(self, spark, log_buffer):
        ch = StreamHandler(log_buffer)
        spark.log.addHandler(ch)
        return ch

    def _get_spark_return_code_from_exception(self, exception):
        match = re.search("Error code is: ([0-9+]).", str(exception))
        if match:
            return match.groups()[0]
        # assume spark_submit success otherwise
        return "0"



# © Copyright Databand.ai, an IBM Company 2022

from dbnd import parameter
from dbnd._core.utils.http.constants import NO_AUTH
from dbnd_spark.spark_config import SparkEngineConfig


class LivySparkConfig(SparkEngineConfig):
    # we don't want spark class to inherit from this one, as it should has Config behaviour
    _conf__task_family = "livy"
    _conf__help_description = "livy configuration"

    url = parameter(
        description="Determine livy's connection url, e.g. `http://livy:8998`"
    )[str]

    auth = parameter(
        description="Set livy auth, e.g. None, Kerberos, or Basic_Access",
        default=NO_AUTH,
    )[str]

    user = parameter(description="Set livy auth user.", default="")[str]

    password = parameter(description="Set livy auth password.", default="")[str]

    ignore_ssl_errors = parameter(
        description="Enable ignoring ssl errors.", default=False
    )[bool]

    job_submitted_hook = parameter(
        description="Set the user code to be run after livy batch submit. This is a reference to a function. "
        "The expected interface is `(LivySparkCtrl, Dict[str, Any]) -> None`",
        default=None,
    )[str]

    job_status_hook = parameter(
        description="Set the user code to be run at each livy batch status update. This is a reference to a function. "
        "The expected interface is `(LivySparkCtrl, Dict[str, Any]) -> None`",
        default=None,
    )[str]

    retry_on_status_error = parameter(
        description="Set the number of retries for http requests if the status code is not accepted.",
        default=20,
    )[int]

    retry_on_status_error_delay = parameter(
        description="Determien the amount of time, in seconds, in between retries for http requests.",
        default=15,
    )[int]

    def get_spark_ctrl(self, task_run):
        from dbnd_spark.livy.livy_spark import LivySparkCtrl

        return LivySparkCtrl(task_run)



# © Copyright Databand.ai, an IBM Company 2022

# do not import PySpark until we run run dbnd_spark_bootstrap
import collections
import logging
import sys
import traceback

from dbnd._core.log import dbnd_log_init_msg
from dbnd._core.utils.seven import import_errors


logger = logging.getLogger(__name__)
_DBND_REGISTER_SPARK_TYPES = None


def dbnd_spark_bootstrap():
    global _DBND_REGISTER_SPARK_TYPES
    if _DBND_REGISTER_SPARK_TYPES:
        return
    # don't run it twice or in recursion
    _DBND_REGISTER_SPARK_TYPES = True

    _workaround_spark_namedtuple_serialization()

    try:
        dbnd_log_init_msg("importing pyspark")

        import pyspark  # noqa: F401
    except import_errors:
        # pyspark is not installed, user will not be able to use pyspark types
        return
    # we register spark types only if we have spark installed
    try:
        from dbnd_spark.spark_targets import dbnd_register_spark_types

        dbnd_register_spark_types()
    except Exception:
        pass
    return


def _workaround_spark_namedtuple_serialization():
    """
    We are checking that we are running in DatProcessor for scheduler,
    DagProcessor will send DAG results back to scheduler via Pipe
    Serialization of DAG with namedtuple will fail because of pyspark patch on namedtuple
    We are going to disable it by "faking" that patch is applied already.
    The moment pyspark will run for real in `airflow run` operator -> this patch is not going to be applied
    and spark will work as usual
    :return:
    """
    if "scheduler" not in sys.argv:
        return

    running_at_scheduler = False
    for filename, line_number, name, line in traceback.extract_stack():
        if line and line.strip() == "self.processor_agent.start()":
            running_at_scheduler = True

    if not running_at_scheduler:
        return

    logging.debug("Preventing pyspark from namedtuple patch")
    setattr(collections.namedtuple, "__hijack", True)
